#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <pcap.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/socket.h>

#include "scan.h"
#include "thread.h"

static void dispatch_callback(u_char *user, const struct pcap_pkthdr *h,
                              const u_char *bytes)
{
	const struct scan *data = (struct scan *)user;
	const struct iphdr *iphdr =
	    (struct iphdr *)(bytes + sizeof(struct ether_header));

	if ((data->type == SCAN_UDP && iphdr->protocol != IPPROTO_UDP) ||
	    (data->type != SCAN_UDP && iphdr->protocol != IPPROTO_TCP)) {
		dprintf(2, "ft_nmap: received packet from a different protocol "
		           "request\n");
		return;
	}

	if (iphdr->protocol == IPPROTO_TCP &&
	    h->caplen >= sizeof(struct ether_header) + sizeof(struct iphdr) +
	                     sizeof(struct tcphdr)) {
		;
	}
	if (iphdr->protocol == IPPROTO_UDP &&
	    h->caplen >= sizeof(struct ether_header) + sizeof(struct iphdr) +
	                     sizeof(struct udphdr)) {
		;
	}
}

int scan(const struct scan *data)
{
	int sockfd = socket(
	    AF_INET, data->type == SCAN_UDP ? SOCK_DGRAM : SOCK_STREAM, 0);
	if (sockfd < 0) {
		dprintf(2, "ft_nmap: failed to create socket");
		return -1;
	}

	char errbuf[PCAP_ERRBUF_SIZE];
	bpf_u_int32 net, mask;

	if (pcap_lookupnet(data->host->interface, &net, &mask, errbuf) < 0) {
		dprintf(2, "ft_nmap: failed to lookup net/mask\n");
		return -1;
	}

	pcap_t *handle =
	    pcap_open_live(data->host->interface, BUFSIZ, 0, 1000, errbuf);
	if (handle == NULL) {
		dprintf(2, "ft_nmap: failed to open device capture\n");
		return -1;
	}

	struct bpf_program fp;
	char pcap_expr[BUFSIZ];
	sprintf(pcap_expr, "src host %s and src port %d and dst host %s",
	        data->dest_addr, data->port, data->host->ip);

	if (pcap_compile(handle, &fp, pcap_expr, 0, mask) < 0) {
		dprintf(2, "ft_nmap: failed to lookup net/mask\n");
		pcap_close(handle);
		return -1;
	}

	if (pcap_setfilter(handle, &fp) < 0) {
		dprintf(2, "ft_nmap: failed to set the filter\n");
		pcap_close(handle);
		return -1;
	}

	pcap_freecode(&fp);

	// send packets

	// TODO test with another cnt value
	if (pcap_dispatch(handle, 10, dispatch_callback, (u_char *)data)) {
		;
	}

	return 0;
}
